/* ==========================================================================
 * libevnet/src/bufio.c - Network server library for libevent.
 * --------------------------------------------------------------------------
 * Copyright (c) 2006  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>	/* NULL */

#include <string.h>	/* strerror(3) memchr(3) memmove(3) */

#include <errno.h>	/* EAGAIN */

#include <assert.h>	/* assert(3) */

#include <unistd.h>	/* read(2) write(3) */

#include <sys/types.h>
#include <sys/param.h>	/* MAX */
#include <sys/queue.h>	/* SLIST_ENTRY SLIST_HEAD SLIST_INIT SLIST_FIRST
			 * SLIST_NEXT */
#include <sys/time.h>	/* struct timeval */

#include <err.h>

#include <event.h>

#include <arena/proto.h>

#include "bufio.h"
#include "tls.h"


#define MARK	fprintf(stderr,"@line %d\n",__LINE__);


#ifndef BUFIO_RECURSE_MAX
#define BUFIO_RECURSE_MAX	(3)
#endif

unsigned int bufio_recurse_max	= BUFIO_RECURSE_MAX;


const char *bufio_errlist[] = {
	[BUFIO_ESUCCESS]		= "Success",
	[BUFIO_ESYSTEM]		= "System error",
	[BUFIO_ETIMEDOUT]		= "Operation timed out",
	[BUFIO_EPENDING]		= "Operation already pending",
	[BUFIO_ECANCELLED]	= "Operation cancelled",
	[BUFIO_ENOMORE]		= "End of IO stream",
	[BUFIO_ENOSPACE]		= "No buffer space available",
	[BUFIO_ENODATA]		= "No more data is available on the write queue",
}; /* bufio_errlist[] */

const int bufio_nerr	= sizeof bufio_errlist / sizeof bufio_errlist[0];



/*
 * During possibly recursive calls keep stack state (or automatic storage)
 * within this structure, which should be pushed onto the stack data queue
 * of the local bufio object. If the bufio object is closed, we can signal this
 * to all the functions that are in scope.
 *
 * This allows information to be relayed safely across the stack and so
 * remove any dependency on the bufio object existing for the entire
 * duration of any running function call (thus allowing them to do useful
 * stuff, like call optimizations, after calling the local callback).
 *
 * Without this we could not implement tail recursion. Note that a
 * longjmp(3) would really throw a monkey wrench into things.
 */
struct bufio_stack_data {
	struct bufio *bufio;

	void *fn_name;

	SLIST_ENTRY(bufio_stack_data) sle;
};


struct bufio_call {
	void *fn_name;

	union {
		struct {
			void *cb_arg;

			struct timeval expire;
		} rdwr;

		struct {
			void *cb_arg;

			struct timeval expire;

			unsigned char *dst;

			size_t dstlen;

			unsigned char *dstpos;

			bufio_read_cb cb;
		} read;

		struct {
			void *cb_arg;

			struct timeval expire;

			const unsigned char *src;

			size_t srclen;

			const unsigned char *srcpos;

			bufio_write_cb cb;
		} write;

		struct {
			void *cb_arg;

			struct timeval expire;
			struct timeval timeout;

			bufio_gets_cb cb;

			struct {
				char *top;
				char *pos;
				char *end;
			} buf;
		} gets;
	} fn_args;
};


static const struct bufio {
	const struct arena_prototype *ap;

	struct event_base *ev_base;

	struct bufio_options opts;

	int fd;

	struct tls *tls;

	enum bufio_errno our_errno;
	int sys_errno;

	struct bufio_type {
		struct bufio_call call;
		struct bufio_call *next; /* Pointer to .call. */

		SLIST_HEAD(, bufio_stack_data) call_stack;
		SLIST_HEAD(, bufio_stack_data) note_stack;

		unsigned int num_calls;

		struct event ev;

		int ev_pending;
		int ev_mode;
	} rd, wr, gl;
} bufio_initializer;


static void bufio_rdwr_try(struct bufio *, struct bufio_type *, int);

static void bufio_read_catch(int fd, short events, void *bufio) {
	if (events & EV_TIMEOUT)
		((struct bufio *)bufio)->rd.ev_pending	= 0;
	/* else check if we set EV_PERSIST */

	bufio_rdwr_try(bufio, &((struct bufio *)bufio)->rd, 0);

	return /* void */;
} /* bufio_read_catch() */


static void bufio_write_catch(int fd, short events, void *bufio) {
	if (events & EV_TIMEOUT)
		((struct bufio *)bufio)->wr.ev_pending	= 0;
	/* else check if we set EV_PERSIST */

	bufio_rdwr_try(bufio, &((struct bufio *)bufio)->wr, 0);

	return /* void */;
} /* bufio_write_catch() */


static void bufio_rdwr_try(struct bufio *bufio, struct bufio_type *t, int queue_only) {
	struct bufio_stack_data sd	= { .bufio = bufio, .fn_name = &bufio_rdwr_try };
	struct bufio_call *cp	= &t->call;
	struct timeval now		= { 0 };
	struct bufio_stack_data *sdp;
	struct timeval timeout;
	enum bufio_errno err;
	enum bufio_mode m;
	ssize_t n;

	/*
	 * Cancel out other calls on this stack.
	 */
	while ((sdp = SLIST_FIRST(&t->note_stack))) {
		SLIST_REMOVE_HEAD(&t->note_stack, sle);

		sdp->bufio	= 0;
	}

	SLIST_INSERT_HEAD(&t->note_stack, &sd, sle);

	if (queue_only) {	/* Fake a blocking call. */
		n	= -1;
		errno	= EAGAIN;
	} else if (t == &bufio->rd) {
		n	= bufio->opts.sys_read(bufio, bufio->fd, cp->fn_args.read.dstpos, &cp->fn_args.read.dst[cp->fn_args.read.dstlen] - cp->fn_args.read.dstpos, cp->fn_args.read.cb_arg);
	} else {
		n	= bufio->opts.sys_write(bufio, bufio->fd, cp->fn_args.write.srcpos, &cp->fn_args.write.src[cp->fn_args.write.srclen] - cp->fn_args.write.srcpos, cp->fn_args.write.cb_arg);
	}

	if (n == -1 && errno == EAGAIN) {
		if (0 != gettimeofday(&now,0)) {
			bufio->sys_errno	= errno;
			bufio->our_errno	= err
					= BUFIO_ESYSTEM;

			m	= BUFIO_ABORT;
		} else if (timercmp(&now, &cp->fn_args.rdwr.expire, >=)) {
			bufio->our_errno	= err
					= BUFIO_ETIMEDOUT;

			m	= BUFIO_ABORT;
		} else {
			/*
			 * Query the IO mode/direction that we need to poll
			 * on before attempting our next IO operation. This
			 * is for protocols like SSL where, in order to
			 * fulfill a logical read, it might need to complete
			 * a low-level write, or vice versa.
			 */
			if (BUFIO_ABORT == (m = bufio->opts.sys_mode(bufio, bufio->fd, ((t == &bufio->rd)? BUFIO_READ : BUFIO_WRITE), t->call.fn_args.rdwr.cb_arg))) {
				bufio->our_errno	= err
						= BUFIO_ECANCELLED;
			}
		}

		n	= 0;
	} else if (n < 0) {
		bufio->sys_errno	= errno;
		bufio->our_errno	= err
				= BUFIO_ESYSTEM;

		n	= 0;
		m	= BUFIO_ABORT;
	} else if (n == 0) {
		bufio->our_errno	= err
				= BUFIO_ENOMORE;

		m	= BUFIO_ABORT;
	} else if (t == &bufio->rd) {
		void *fn_name	= cp->fn_name;
		void *pos	= cp->fn_args.read.dstpos;
		int again;

		cp->fn_name		= 0;
		cp->fn_args.read.dstpos	+= n;

		if ((again = cp->fn_args.read.cb(sd.bufio, pos, n, BUFIO_ESUCCESS, cp->fn_args.read.cb_arg))) {
			assert(sd.bufio && !t->call.fn_name);

			/* Restart IO. */
			cp->fn_name	= fn_name;
			n		= 0;

			if (cp->fn_args.read.dstpos >= &cp->fn_args.read.dst[cp->fn_args.read.dstlen]) {
				m	= BUFIO_ABORT;

				bufio->our_errno	= err
							= BUFIO_ENOSPACE;
			} else if (BUFIO_ABORT == (m = bufio->opts.sys_mode(bufio, bufio->fd, BUFIO_READ, t->call.fn_args.read.cb_arg))) {
				bufio->our_errno	= err
							= BUFIO_ECANCELLED;
			}
		} else if (sd.bufio) {
			/* Do not allow another callback. */
			cp	= 0;

			/* Remove the event. */
			m	= BUFIO_ABORT;
		} else
			goto done;
	} else if (t == &bufio->wr) {
		void *fn_name	= cp->fn_name;
		const void *pos	= cp->fn_args.write.srcpos;
		int again;

		cp->fn_name			= 0;
		cp->fn_args.write.srcpos	+= n;

		if ((again = cp->fn_args.write.cb(sd.bufio, pos, n, BUFIO_ESUCCESS, cp->fn_args.write.cb_arg))) {
			assert(sd.bufio && !sd.bufio->wr.call.fn_name);

			/* Restart IO. */
			cp->fn_name	= fn_name;
			n		= 0;

			if (cp->fn_args.write.srcpos >= &cp->fn_args.write.src[cp->fn_args.write.srclen]) {
				m	= BUFIO_ABORT;

				bufio->our_errno	= err
						= BUFIO_ENODATA;
			} else if (BUFIO_ABORT == (m = bufio->opts.sys_mode(bufio, bufio->fd, BUFIO_WRITE, t->call.fn_args.rdwr.cb_arg))) {
				bufio->our_errno	= err
						= BUFIO_ECANCELLED;
			}
		} else if (sd.bufio) {
			/* Do not allow another callback. */
			cp	= 0;

			/* Remove the event. */
			m	= BUFIO_ABORT;
		} else
			goto done;
	} else {
		assert(0 && (t == &bufio->rd || t == &bufio->wr));
	}

	switch (m) {
	case BUFIO_READ:
		/* FALL THROUGH */
	case BUFIO_WRITE:
		if (t->ev_pending) {
			if (t->ev_mode == m)
				break;
 
			(void)event_del(&t->ev);

			t->ev_pending	= 0;
		}

		event_set(&t->ev, sd.bufio->fd, ((m == BUFIO_READ)? EV_READ : EV_WRITE) | EV_PERSIST, (t == &sd.bufio->rd)? &bufio_read_catch : &bufio_write_catch, bufio);

		if (sd.bufio->ev_base)
			event_base_set(sd.bufio->ev_base,&t->ev);

		if (0 != gettimeofday(&now,0)) {
			bufio->sys_errno	= errno;
			bufio->our_errno	= err
					= BUFIO_ESYSTEM;

			goto abort;
		}

		if (timercmp(&now, &cp->fn_args.rdwr.expire, >))
			cp->fn_args.rdwr.expire	= now;

		timersub(&cp->fn_args.rdwr.expire,&now,&timeout);

		if (0 != event_add(&t->ev,&timeout)) {
			bufio->sys_errno	= errno;
			bufio->our_errno	= err
					= BUFIO_ESYSTEM;

			goto abort;
		}

		t->ev_mode	= m;
		t->ev_pending	= 1;

		break;
	case BUFIO_ABORT:
abort:
		if (t->ev_pending) {
			/* Cannot check error here because epoll(2) and
			 * others might return an error if the descriptor
			 * was closed and had already implicitly dequeued
			 * the event. Such is a benign circumstance, and
			 * very common in some situations.
			 */
			(void)event_del(&t->ev);

			t->ev_pending	= 0;
		}

		if (cp) {
			if (t == &sd.bufio->rd) {
				assert(0 == cp->fn_args.read.cb(sd.bufio, cp->fn_args.read.dstpos, n, err, cp->fn_args.read.cb_arg));
			} else {
				assert(0 == cp->fn_args.write.cb(sd.bufio, cp->fn_args.write.srcpos, n, err, cp->fn_args.write.cb_arg));
			}
		}

		break;
	default:
		assert(!"Bad IO mode");
	}

done:
	if (sd.bufio) {
		assert(&sd == SLIST_FIRST(&t->note_stack));

		SLIST_REMOVE_HEAD(&t->note_stack, sle);
	}

	return /* void */;
} /* bufio_rdwr_try() */


enum bufio_errno bufio_read(struct bufio *bufio, void *dst, size_t dstlen, bufio_read_cb cb, void *arg, struct timeval *timeout) {
	struct bufio_stack_data sd	= { .bufio = bufio, .fn_name = &bufio_read };
	enum bufio_errno err		= BUFIO_ESUCCESS;
	struct timeval now;

	if (SLIST_EMPTY(&sd.bufio->rd.call_stack))
		sd.bufio->rd.num_calls	= 0;

	SLIST_INSERT_HEAD(&sd.bufio->rd.call_stack, &sd, sle);

	if (sd.bufio->rd.call.fn_name) {
		sd.bufio->our_errno	= err
					= BUFIO_EPENDING;

		goto done;
	}

	if (0 != gettimeofday(&now,0)) {
		sd.bufio->sys_errno	= errno;
		sd.bufio->our_errno	= err
					= BUFIO_ESYSTEM;

		goto done;
	}

	assert(!sd.bufio->rd.next);

	sd.bufio->rd.call.fn_name		= sd.fn_name;
	sd.bufio->rd.call.fn_args.read.dst	= dst;
	sd.bufio->rd.call.fn_args.read.dstlen	= dstlen;
	sd.bufio->rd.call.fn_args.read.dstpos	= dst;
	sd.bufio->rd.call.fn_args.read.cb	= cb;
	sd.bufio->rd.call.fn_args.read.cb_arg	= arg;

	timeradd(&now,timeout,&sd.bufio->rd.call.fn_args.read.expire);

	sd.bufio->rd.next	= &sd.bufio->rd.call;

#if 1	/* Pass down the stack? */
	if (SLIST_NEXT(&sd, sle) && SLIST_NEXT(&sd, sle)->fn_name == sd.fn_name)
		goto done;
#endif

	while (sd.bufio && sd.bufio->rd.next && sd.bufio->rd.next->fn_name == sd.fn_name) {
		sd.bufio->rd.next	= 0;	/* Unset; Just pointer to .call. */

		bufio_rdwr_try(sd.bufio, &sd.bufio->rd, (++sd.bufio->rd.num_calls > sd.bufio->opts.recurse_max));
	}

done:
	if (sd.bufio) {
		assert(&sd == SLIST_FIRST(&sd.bufio->rd.call_stack));

		SLIST_REMOVE_HEAD(&sd.bufio->rd.call_stack, sle);
	}

	return err;
} /* bufio_read() */


enum bufio_errno bufio_write(struct bufio *bufio, const void *src, size_t srclen, bufio_write_cb cb, void *arg, struct timeval *timeout) {
	struct bufio_stack_data sd	= { .bufio = bufio, .fn_name = &bufio_write };
	enum bufio_errno err		= BUFIO_ESUCCESS;
	struct timeval now;

	if (SLIST_EMPTY(&sd.bufio->wr.call_stack))
		sd.bufio->wr.num_calls	= 0;

	SLIST_INSERT_HEAD(&sd.bufio->wr.call_stack, &sd, sle);

	if (sd.bufio->wr.call.fn_name) {
		sd.bufio->our_errno	= err
					= BUFIO_EPENDING;

		goto done;
	}

	if (0 != gettimeofday(&now,0)) {
		sd.bufio->sys_errno	= errno;
		sd.bufio->our_errno	= err
					= BUFIO_ESYSTEM;

		goto done;
	}

	assert(!sd.bufio->wr.next);

	sd.bufio->wr.call.fn_name			= sd.fn_name;
	sd.bufio->wr.call.fn_args.write.src	= src;
	sd.bufio->wr.call.fn_args.write.srclen	= srclen;
	sd.bufio->wr.call.fn_args.write.srcpos	= src;
	sd.bufio->wr.call.fn_args.write.cb	= cb;
	sd.bufio->wr.call.fn_args.write.cb_arg	= arg;

	timeradd(&now,timeout,&sd.bufio->wr.call.fn_args.read.expire);

	sd.bufio->wr.next	= &sd.bufio->wr.call;

#if 1	/* Pass down the stack? */
	if (SLIST_NEXT(&sd, sle) && SLIST_NEXT(&sd, sle)->fn_name == sd.fn_name)
		goto done;
#endif

	while (sd.bufio && sd.bufio->wr.next && sd.bufio->wr.next->fn_name == sd.fn_name) {
		sd.bufio->wr.next	= 0;	/* Unset; Just pointer to .call. */

		bufio_rdwr_try(sd.bufio, &sd.bufio->wr, (++sd.bufio->wr.num_calls > sd.bufio->opts.recurse_max));
	}

done:
	if (sd.bufio) {
		assert(&sd == SLIST_FIRST(&sd.bufio->wr.call_stack));

		SLIST_REMOVE_HEAD(&sd.bufio->wr.call_stack, sle);
	}

	return err;
} /* bufio_write() */


static int bufio_gets_tell(struct bufio *bufio, char *eol, enum bufio_errno err) {
	struct bufio_stack_data sd	= { .bufio = bufio, .fn_name = &bufio_gets_tell };
	struct bufio_call *cp	= &bufio->gl.call;
	int again, ch;

	SLIST_INSERT_HEAD(&sd.bufio->gl.note_stack, &sd, sle);

	if (cp->fn_args.gets.buf.pos > cp->fn_args.gets.buf.top) {
		assert(eol >= cp->fn_args.gets.buf.top && eol < cp->fn_args.gets.buf.end);

		ch	= *eol;
		*eol	= '\0';
	}

	again	= cp->fn_args.gets.cb(bufio, cp->fn_args.gets.buf.top, eol - cp->fn_args.gets.buf.top, err, cp->fn_args.gets.cb_arg);

	if (sd.bufio) {
		if (cp->fn_args.gets.buf.pos > cp->fn_args.gets.buf.top) {
			assert(eol >= cp->fn_args.gets.buf.top && eol < cp->fn_args.gets.buf.end);

			*eol	= ch;
		}

		(void)memmove(cp->fn_args.gets.buf.top, eol, cp->fn_args.gets.buf.pos - eol);

		cp->fn_args.gets.buf.pos	= cp->fn_args.gets.buf.top + (cp->fn_args.gets.buf.pos - eol);
	}

	if (sd.bufio) {
		assert(&sd == SLIST_FIRST(&sd.bufio->gl.note_stack));

		SLIST_REMOVE_HEAD(&sd.bufio->gl.note_stack, sle);
	}

	return again;
} /* bufio_gets_tell() */


static enum bufio_action bufio_gets_catch(struct bufio *bufio, void *bytes, size_t nbytes, enum bufio_errno err, void *nul) {
	struct bufio_call *cp	= &bufio->gl.call;

	if (err != BUFIO_ESUCCESS) {
		cp->fn_name	= 0;

		if (err == BUFIO_ENOMORE && cp->fn_args.gets.buf.pos > cp->fn_args.gets.buf.top) {
			assert(0 == bufio_gets_tell(bufio, cp->fn_args.gets.buf.pos, err));
		} else {
			assert(0 == cp->fn_args.gets.cb(bufio, 0, 0, err, cp->fn_args.gets.cb_arg));
		}

		return 0;
	}

	cp->fn_args.gets.buf.pos	+= nbytes;

	/*
	 * Call back into the public interface as-if starting from scratch
	 * so that we can trigger call coalescing.
	 */
	cp->fn_name	= 0;	/* Cheat. This is stupid, BTW. */
	bufio_gets(bufio, cp->fn_args.gets.cb, cp->fn_args.gets.cb_arg, &cp->fn_args.gets.timeout);

	return 0;
} /* bufio_gets_catch() */


static void bufio_gets_try(struct bufio *bufio) {
	struct bufio_stack_data sd	= { .bufio = bufio, .fn_name = &bufio_gets_try };
	struct bufio_call *cp	= &sd.bufio->gl.call;
	void *fn_name			= cp->fn_name;
	int again			= 1;
	char *nl;

	SLIST_INSERT_HEAD(&sd.bufio->gl.note_stack, &sd, sle);

	while (again && (nl = memchr(cp->fn_args.gets.buf.top, '\n', cp->fn_args.gets.buf.pos - cp->fn_args.gets.buf.top))) {
		assert(sd.bufio);

		cp->fn_name	= 0;
		again		= bufio_gets_tell(bufio, nl + 1, BUFIO_ESUCCESS);
	} /* while (again && nl) */

	if (again) {
		assert(sd.bufio);

		cp->fn_name	= fn_name;

		if (0 == (cp->fn_args.gets.buf.end - cp->fn_args.gets.buf.pos)) {
			size_t want	= MAX(512, 2 * (cp->fn_args.gets.buf.end - cp->fn_args.gets.buf.top));
			char *p;

			if (!(p = (bufio->ap->realloc)(bufio->ap, cp->fn_args.gets.buf.top, want, 1)))
				goto sysfail;

			cp->fn_args.gets.buf.end	= p + want;
			cp->fn_args.gets.buf.pos	= p + (cp->fn_args.gets.buf.pos - cp->fn_args.gets.buf.top);
			cp->fn_args.gets.buf.top	= p;
		}

		if (0 != bufio_read(bufio, cp->fn_args.gets.buf.pos, (cp->fn_args.gets.buf.end - cp->fn_args.gets.buf.pos) - 1, &bufio_gets_catch, 0, &cp->fn_args.gets.timeout))
				goto bufiofail;
	}

	goto done;
sysfail:
	bufio->sys_errno	= errno;
	bufio->our_errno	= BUFIO_ESYSTEM;

	/* FALL THROUGH */
bufiofail:
	cp->fn_name	= 0;

	assert(0 == cp->fn_args.gets.cb(bufio, 0, 0, bufio->our_errno, cp->fn_args.gets.cb_arg));

	/* FALL THROUGH */
done:
	if (sd.bufio) {
		assert(&sd == SLIST_FIRST(&sd.bufio->gl.note_stack));

		SLIST_REMOVE_HEAD(&sd.bufio->gl.note_stack, sle);
	}

	return /* void */;
} /* bufio_gets_try() */


enum bufio_errno bufio_gets(struct bufio *bufio, bufio_gets_cb cb, void *arg, struct timeval *timeout) {
	struct bufio_stack_data sd	= { .bufio = bufio, .fn_name = &bufio_gets };
	enum bufio_errno err		= BUFIO_ESUCCESS;
	struct timeval now;

	if (SLIST_EMPTY(&sd.bufio->gl.call_stack))
		sd.bufio->gl.num_calls	= 0;

	SLIST_INSERT_HEAD(&sd.bufio->gl.call_stack, &sd, sle);

	if (sd.bufio->gl.call.fn_name) {
		sd.bufio->our_errno	= err
					= BUFIO_EPENDING;
assert(0);
		goto done;
	}

	if (0 != gettimeofday(&now,0)) {
		sd.bufio->sys_errno	= errno;
		sd.bufio->our_errno	= err
					= BUFIO_ESYSTEM;

		goto done;
	}

	assert(!sd.bufio->gl.next);

	sd.bufio->gl.call.fn_name		= sd.fn_name;
	sd.bufio->gl.call.fn_args.gets.cb	= cb;
	sd.bufio->gl.call.fn_args.gets.cb_arg	= arg;

	if (timeout) {
		sd.bufio->gl.call.fn_args.gets.timeout	= *timeout;

		timeradd(&now,timeout,&sd.bufio->gl.call.fn_args.gets.expire);
	}

	sd.bufio->gl.next	= &sd.bufio->gl.call;

	/* Pass down the stack? */
	if (SLIST_NEXT(&sd, sle) && SLIST_NEXT(&sd, sle)->fn_name == sd.fn_name)
		goto done;

	while (sd.bufio && sd.bufio->gl.next && sd.bufio->gl.next->fn_name == sd.fn_name) {
		sd.bufio->gl.next	= 0;	/* Unset; Just pointer to .call. */

		sd.bufio->gl.num_calls++;

		bufio_gets_try(sd.bufio);
	}

done:
	if (sd.bufio) {
		assert(&sd == SLIST_FIRST(&sd.bufio->gl.call_stack));

		SLIST_REMOVE_HEAD(&sd.bufio->gl.call_stack, sle);
	}

	return err;
} /* bufio_gets() */


static enum bufio_mode bufio_sys_mode(struct bufio *bufio, int fd, enum bufio_mode want, void *null) {
	return want;
} /* bufio_sys_mode() */


static ssize_t bufio_sys_read(struct bufio *bufio, int fd, void *p, size_t n, void *null) {
	return read(fd,p,n);
} /* bufio_sys_read() */


static ssize_t bufio_sys_write(struct bufio *bufio, int fd, const void *p, size_t n, void *null) {
	return write(fd,p,n);
} /* bufio_sys_write() */


const struct bufio_options bufio_defaults = {
	.recurse_max	= -1,
	.sys_mode	= &bufio_sys_mode,
	.sys_read	= &bufio_sys_read,
	.sys_write	= &bufio_sys_write,
};


struct bufio *bufio_open(int fd, const struct bufio_options *opts, struct event_base *ev_base, const struct arena_prototype *ap) {
	struct bufio *bufio;

	if (!opts)
		opts	= &bufio_defaults;

	if (!ap)
		ap	= ARENA_STDLIB;

	if (!(bufio = ap->malloc(ap, sizeof *bufio, 0)))
		return 0;

	*bufio	= bufio_initializer;

	bufio->ap		= ap;
	bufio->ev_base	= ev_base;
	bufio->opts	= *opts;
	bufio->fd		= fd;

	SLIST_INIT(&bufio->rd.call_stack);
	SLIST_INIT(&bufio->rd.note_stack);

	SLIST_INIT(&bufio->wr.call_stack);
	SLIST_INIT(&bufio->wr.note_stack);

	SLIST_INIT(&bufio->gl.call_stack);
	SLIST_INIT(&bufio->gl.note_stack);

	bufio->gl.call.fn_args.gets.buf.top	= 0;
	bufio->gl.call.fn_args.gets.buf.pos	= 0;
	bufio->gl.call.fn_args.gets.buf.end	= 0;

	if (bufio->opts.recurse_max == -1)
		bufio->opts.recurse_max	= bufio_recurse_max;

	return bufio;
} /* bufio_open() */


enum bufio_errno bufio_close(struct bufio *bufio) {
	struct bufio_stack_data *sd;

	/* TODO: Allow early termination. */
	if (bufio->wr.ev_pending || bufio->rd.ev_pending)
		return BUFIO_EPENDING;

	/*
	 * Cleanup our call stack so we don't do stupid stuff (and if we do
	 * we [hopefully] immediately segfault).
	 */
	while ((sd = SLIST_FIRST(&bufio->rd.call_stack))) {
		sd->bufio	= 0;

		SLIST_REMOVE_HEAD(&bufio->rd.call_stack, sle);
	}

	while ((sd = SLIST_FIRST(&bufio->rd.note_stack))) {
		sd->bufio	= 0;

		SLIST_REMOVE_HEAD(&bufio->rd.note_stack, sle);
	}

	while ((sd = SLIST_FIRST(&bufio->wr.call_stack))) {
		sd->bufio	= 0;

		SLIST_REMOVE_HEAD(&bufio->wr.call_stack, sle);
	}

	while ((sd = SLIST_FIRST(&bufio->wr.note_stack))) {
		sd->bufio	= 0;

		SLIST_REMOVE_HEAD(&bufio->wr.note_stack, sle);
	}

	while ((sd = SLIST_FIRST(&bufio->gl.call_stack))) {
		sd->bufio	= 0;

		SLIST_REMOVE_HEAD(&bufio->gl.call_stack, sle);
	}

	while ((sd = SLIST_FIRST(&bufio->gl.note_stack))) {
		sd->bufio	= 0;

		SLIST_REMOVE_HEAD(&bufio->gl.note_stack, sle);
	}

	bufio->ap->free(bufio->ap,bufio);

	return BUFIO_ESUCCESS;
} /* bufio_close() */


static enum bufio_mode bufio_tls_mode(struct bufio *bufio, int fd, enum bufio_mode want, void *null) {
	switch ((TLS_S_NEED_READ|TLS_S_NEED_WRITE) & tls_state(bufio->tls)) {
	case TLS_S_NEED_READ|TLS_S_NEED_WRITE:
		/* FALL THROUGH */
	case TLS_S_NEED_READ:
		return BUFIO_READ;
	case TLS_S_NEED_WRITE:
		return BUFIO_WRITE;
	default:
		return want;
	}

	/* NOT REACHED */
} /* bufio_sys_mode() */


static ssize_t bufio_tls_read(struct bufio *bufio, int fd, void *p, size_t n, void *null) {
	return tls_read(bufio->tls, p, n, 0, 0, 0);
} /* bufio_sys_read() */


static ssize_t bufio_tls_write(struct bufio *bufio, int fd, const void *p, size_t n, void *null) {
	return tls_write(bufio->tls, p, n, 0, 0, 0);
} /* bufio_sys_write() */


void bufio_tls_start(struct bufio *bufio, struct tls *tls) {
	bufio->tls	= tls;

	bufio->opts.sys_mode	= &bufio_tls_mode;
	bufio->opts.sys_read	= &bufio_tls_read;
	bufio->opts.sys_write	= &bufio_tls_write;

	return /* void */;
} /* bufio_tls_start() */


int bufio_sys_errno(struct bufio *bufio) {
	return bufio->sys_errno;
} /* bufio_sys_errno() */


enum bufio_errno bufio_errno(struct bufio *bufio) {
	return bufio->our_errno;
} /* bufio_errno() */


const char *bufio_strerror(struct bufio *bufio) {
	if (bufio->our_errno == BUFIO_ESYSTEM)
		return strerror(bufio->sys_errno);
	else
		return bufio_errlist[bufio->our_errno];
} /* bufio_strerror() */


